module Gale(GaleContext,Puff(..),galeNextPuff,showPuff, galeDecryptPuff, newGaleContext, Fragment(..), Category,
    catParseNew, catParseOld, catShowNew, catShowOld, subCategory, destroyGaleContext, connectGaleContext, reconnectGaleContext,
    getFragmentString, getFragmentStrings, writePuffs, readPuffs) where


import Network
import Concurrent
import ConcurrentUtil
import IO
import IOExts
import Char(chr,ord)
import Word
import Int
import Bits
import GenUtil
import List
import Monad(when)
import Time
import Control.Exception
import Bits


-- TODO - prove concurrent-correctness, make sure all network errors are accounted for.

galePort :: PortNumber
galePort = 11512
hostStrings :: String -> [String]
hostStrings s = [s, "gale." ++ s, s ++ ".gale.org."]


data GaleContext = GaleContext { channel :: Chan Puff, categories :: [(String,String)], threads :: MVar [ThreadId] }

type Category = (String,String)
    
data Puff = Puff [Category] [(String,Fragment)] | Message String 
data Fragment = FragmentText String | FragmentData [Word8] | FragmentTime ClockTime | FragmentInt Int32 | FragmentNest [(String,Fragment)]




subCategory :: Category -> Category -> Bool
(x1,y1) `subCategory` (x2,y2) = x2 `isPrefixOf` x1 && y2 `isSuffixOf` y1

tryMap :: (a -> b) -> [a] -> IO [b]
tryMap f xs = do 
    ys <- mapM (tryMost . evaluate . f ) xs
    return [y|(Right y) <- ys]
    

newGaleContext cs = do
    cs' <- tryMap catParseNew cs
    c <- newChan 
    mv <- newEmptyMVar
    let gc = GaleContext { channel = c, categories = cs', threads = mv}
    connectGaleContext gc
    return gc

spc [] = []
spc s = v : spc (drop 1 r) where
    (v,r) = span (/= ':') s 

sendEmptyPuff h = do
    putWord32 h 0
    putWord32 h 8
    putWord32 h 0
    putWord32 h 0

passKilled (AsyncException ThreadKilled) = Nothing
passKilled x = Just x

catchMost = Control.Exception.catchJust passKilled

tryMost = Control.Exception.tryJust passKilled

connectThread c cs d  = do
    let continue = threadDelay 2000000 >> connectThread c cs d
    h' <- tryMost $ trySeveral (map attemptConnect (hostStrings d))
    case h' of 
	Left _ -> continue
	Right h -> catchMost  (do
	    putWord32 h 1
	    v <- readWord32 h
	    let ss = nub [c| (c,d') <- cs, d' == d]
		f c = catShowOld (c,d)
		gs = concatInter ":" (map f ss)
	    writeChan c (Message $ "Connected: " ++  d ++ " -- " ++ unwords ss)
	    putWord32 h 2
	    putWord32 h (fromIntegral $ length gs * 2)
	    putRaw h $ galeEncodeString gs
	    repeatM_ $ do
		w <- readWord32 h
		l <- readWord32 h
		d <- readRaw h (fromIntegral l)
		when (w == 0) $ do
		    let (l,d') = xdrReadUInt d
		    writeChan c $ Puff (map catParseOld $ spc $ galeDecodeString (take (fromIntegral $ l ) d'))  (decodeFragments $ drop (fromIntegral l + 4) d'))
	      $ \_ -> do 
		hClose h
		writeChan c (Message $ "Disconnected: " ++ d)
		continue

destroyGaleContext :: GaleContext -> IO ()
destroyGaleContext gc = do
    thds <- takeMVar (threads gc)
    mapM_ killThread thds
    

reconnectGaleContext gc = closeGaleContext gc >> connectGaleContext gc

closeGaleContext = destroyGaleContext
connectGaleContext gc = do
    thds <- mapM (forkIO . connectThread (channel gc) (categories gc)) (nub $ map snd (categories gc))
    putMVar (threads gc) thds

    
galeNextPuff :: GaleContext -> IO Puff
galeNextPuff gc = readChan $ channel gc

showPuff (Puff s fl) = unwords (map catShowNew s) ++ " \n" ++ concatMap (\(n,f) -> n ++ ": " ++ showFragment f ++ "\n") fl
showPuff (Message s) = s

showFragment (FragmentText s) = s
showFragment (FragmentData _) = "<DATA>"
showFragment (FragmentTime ct) = show ct
showFragment (FragmentInt x) = show x
showFragment (FragmentNest xs) = concatMap (\(n,f) -> n ++ ": " ++ showFragment f ++ "\n") xs

galeDecryptPuff :: Puff -> IO Puff 
galeDecryptPuff (Puff s [("security/signature",FragmentData xs)]) = return $ Puff s fl where
    (l,xs') = xdrReadUInt xs
    fl = (decodeFragments $ drop (fromIntegral l + 4) xs')
galeDecryptPuff x = return x
    
decodeFragments :: [Word8] -> [(String,Fragment)]
decodeFragments [] = []
decodeFragments xs = decodeFragment ft f : decodeFragments r where
    (ft,xs') = xdrReadUInt xs
    (fl,xs'') = xdrReadUInt xs'
    (f,r) = splitAt (fromIntegral fl) xs''

--decodeFragment ft f = (show ft ++ " " ++ show f,FragmentTime)
decodeFragment :: Word32 -> [Word8] -> (String, Fragment)
decodeFragment t xs = (fn,f t) where
    (fnl,xs') = xdrReadUInt xs
    (fn,xs'') = liftT2 (galeDecodeString,id) (splitAt (fromIntegral $ fnl * 2) xs')
    f 0 = FragmentText (galeDecodeString xs'')
    f 1 = FragmentData xs''
    f 2 = FragmentTime $ decodeTime xs''
    f 3 = FragmentInt (fromIntegral $ fst $ xdrReadUInt xs'')
    f 4 = FragmentNest (decodeFragments xs'')

decodeTime :: [Word8] -> ClockTime
decodeTime xs = TimeDiff {tdDay = 0, tdPicosec = 0, tdYear = 0, tdMonth = 0, tdHour = 0, tdMin = fromIntegral (t `div` 60) , tdSec = fromIntegral (t `mod` 60)} `addToClockTime` epoch where
    (t,_) = xdrReadUInt (drop 4 xs)

attemptConnect s = do
    h <- connectTo s (PortNumber galePort)
    hSetBuffering h NoBuffering
    return h


putRaw :: Handle -> [Word8] -> IO ()
putRaw h v = hPutStr h (map (chr . fromIntegral) v)

readRaw :: Handle -> Int -> IO [Word8]
readRaw h n = do 
    v <- replicateM n (hGetChar h)
    return $ map (fromIntegral . ord) v

type XDRShowS = [Word8] -> [Word8]
type XDRReadS a = [Word8] -> (a,[Word8])


xdrWriteUInt :: Word32 -> XDRShowS
xdrWriteUInt x bs = (b1:b2:b3:b4:bs) where
    b1 = fromIntegral $ (x `shiftR` 24) .&. 0xFF
    b2 = fromIntegral $ (x `shiftR` 16) .&. 0xFF
    b3 = fromIntegral $ (x `shiftR` 8) .&. 0xFF
    b4 = fromIntegral $ x .&. 0xFF

xdrReadUInt :: XDRReadS Word32
xdrReadUInt (b1:b2:b3:b4:bs) = (x,bs) where
    x = (fromIntegral b4) .|. (fromIntegral b3 `shiftL` 8) .|. 
	    (fromIntegral b2 `shiftL` 16) .|. (fromIntegral b1 `shiftL` 24)
xdrReadUInt bs = error $ "xdrReadUInt " ++ show bs


putWord32 h x = do
    putRaw h $ xdrWriteUInt x []

readWord32 h = do
    xs <- readRaw h 4 
    return $ fst $ xdrReadUInt xs

galeEncodeString :: String -> [Word8]
galeEncodeString cs = concatMap (f . ord) cs where
    f x = (b1:b2:[]) where 
	b1 = fromIntegral $ (x `shiftR` 8) .&. 0xFF
	b2 = fromIntegral $ x .&. 0xFF

galeDecodeString :: [Word8] -> String
galeDecodeString [] = [] 
galeDecodeString (b1:b2:xs) = (chr $ (fromIntegral b2) .|. (fromIntegral b1 `shiftL` 8)) : galeDecodeString xs

--------------------
-- Category routines
--------------------

catfixes = [ ("/", ".|"), (".", "/"), (":", "..") ]

categoryO2N ('@':r) = con (init c) ++ "@" ++ d where
    (d,'/':'u':'s':'e':'r':'/':c) = span (/= '/') r
    con cs | [nv] <- [x ++ (con $ drop (length y) cs) |(x,y) <- catfixes, y `isPrefixOf` cs] = nv
    con (c:cs) = c:con cs
    con "" = "" 
    
catParseNew :: String -> Category
catParseNew cs = (c,tail d) where
    (c,d) = span (/= '@') cs
    
catParseOld :: String -> Category
catParseOld ('@':r) = (con (init c),d) where
    (d,'/':'u':'s':'e':'r':'/':c) = span (/= '/') r
    con cs | [nv] <- [x ++ (con $ drop (length y) cs) |(x,y) <- catfixes, y `isPrefixOf` cs] = nv
    con (c:cs) = c:con cs
    con "" = "" 

catShowOld :: Category -> String
catShowOld (c,d) = "@" ++ d ++ "/user/" ++ con c where 
    con cs | [nv] <- [x ++ (con $ drop (length y) cs) |(y,x) <- catfixes, y `isPrefixOf` cs] = nv
    con (c:cs) = c:con cs
    con "" = "" 

catShowNew :: Category -> String
catShowNew (x,y) = x ++ "@" ++ y

------------------
-- Examining puffs
------------------

getFragmentString :: Puff -> String -> Maybe String
getFragmentString (Puff _ frags) s = case lookup s frags of
    Just (FragmentText s) -> Just s
    _ -> Nothing

getFragmentStrings :: Puff -> String -> [String]
getFragmentStrings (Puff _ frags) s = [v|(n,FragmentText v) <- frags, n == s]


----------------
-- Puff Pickling
----------------

newtype PickledPuffs = PickledPuffs [Puff]

-- data Puff = Puff [Category] [(String,Fragment)] | Message String 
-- data Fragment = FragmentText String | FragmentData [Word8] | FragmentTime ClockTime | FragmentInt Int32 | FragmentNest [(String,Fragment)]


writePuffs :: String -> [Puff] -> IO ()
writePuffs fn ps = do
    bh <- newBinHandle []
    put bh ps
    cs <- readBinHandle bh
    tryMost (writeFile fn cs) 
    return ()

readPuffs :: String -> IO [Puff]
readPuffs fn = do
    v <- tryMost (readFile fn)
    case v of
	Left _ -> return []
	Right v -> do
		bh <- newBinHandle v
		tryMost (get bh) >>= \r -> case r of
		    Left _ -> return []
		    Right ps -> return ps




type BinHandle = (IORef [Char])
type Bin a = Int

newBinHandle x = newIORef (reverse x)
readBinHandle bh =  readIORef bh


putBits :: BinHandle -> Int -> Int -> IO (Bin a)
putBits _ 0 _ = return 0
putBits bh nbits val = do
    modifyIORef bh ("01" !! (val `mod` 2):)
    putBits bh (nbits - 1) (val `shiftR` 1)
getBits :: BinHandle -> Int -> IO Int
getBits bh nbits = do
    s <- readIORef bh
    modifyIORef bh (drop nbits)
    let cs = map f $ reverse (take nbits s) 
    return $ foldl (\c x -> (c `shiftL` 1) + x ) 0 cs
    
    where
	f '0' = 0
	f '1' = 1
    



class Binary a where
    put    :: BinHandle -> a -> IO (Bin a)
    get    :: BinHandle -> IO a


instance Binary () where
    put  bh () = putBits bh 0 0
    get  _     = return ()

instance Binary Bool where
    put  bh b = putBits bh 1 (fromEnum b)
    get  bh   = getBits bh 1 >>= return . toEnum

instance Binary Char where
    put  bh c = putBits bh 8 (fromEnum c)
    get  bh   = do x <- getBits bh 8; return (toEnum x)

instance Binary Int where
    put  bh i = putBits  bh 32 i
    get  bh   = getBits  bh 32

instance Binary Word8 where
    put  bh i = putBits bh 8 (fromIntegral i)
    get  bh   = fmap fromIntegral $  getBits bh 8

instance Binary Int32 where
    put  bh i = putBits bh 32 (fromIntegral i)
    get  bh   = fmap fromIntegral $  getBits bh 32


instance Binary ClockTime where
    put bh ct = do
	let t = toUTCTime ct
	put bh (ctYear t)
	put bh (fromEnum $ ctMonth t)
	put bh (ctDay t)
	put bh (ctHour t)
	put bh (ctMin t)
	put bh (ctSec t)
    get bh = do
	year <- get bh
	month <- fmap toEnum $ get bh 
	day <- get bh 
	hour <- get bh 
	min <- get bh 
	sec <- get bh 
	return $ toClockTime $ (toUTCTime epoch) {ctYear = year, ctDay = day, ctMonth = month, ctHour = hour, ctMin = min, ctSec = sec}
	

instance (Binary a, Binary b) => Binary (a,b) where
    put bh (a,b) = do pos <- putBits bh 0 0
                      put bh a
                      put bh b
                      return pos
    get bh       = do a <- get bh
                      b <- get bh
                      return (a,b)

instance Binary a => Binary [a] where
    put bh []     = putBits bh 1 0
    put bh (x:xs) = do pos <- putBits bh 1 1
                       put bh x
                       put bh xs
                       return pos
    get bh        = do h <- getBits bh 1
                       case h of
                          0 -> return []
                          _ -> do x  <- get bh
                                  xs <- get bh
                                  return (x:xs)

instance Binary Puff where
    put bh (Puff aa ab) = do
	    pos <- putBits bh 1 0
	    put bh aa
	    put bh ab
	    return pos
    put bh (Message ac) = do
	    pos <- putBits bh 1 1
	    put bh ac
	    return pos
    get bh = do
	    h <- getBits bh 1
	    case h of
	      0 -> do
		    aa <- get bh
		    ab <- get bh
		    return (Puff aa ab)
	      1 -> do
		    ac <- get bh
		    return (Message ac)

instance Binary Fragment where
    put bh (FragmentText aa) = do
	    pos <- putBits bh 3 0
	    put bh aa
	    return pos
    put bh (FragmentData ab) = do
	    pos <- putBits bh 3 1
	    put bh ab
	    return pos
    put bh (FragmentTime ac) = do
	    pos <- putBits bh 3 2
	    put bh ac
	    return pos
    put bh (FragmentInt ad) = do
	    pos <- putBits bh 3 3
	    put bh ad
	    return pos
    put bh (FragmentNest ae) = do
	    pos <- putBits bh 3 4
	    put bh ae
	    return pos
    get bh = do
	    h <- getBits bh 3
	    case h of
	      0 -> do
		    aa <- get bh
		    return (FragmentText aa)
	      1 -> do
		    ab <- get bh
		    return (FragmentData ab)
	      2 -> do
		    ac <- get bh
		    return (FragmentTime ac)
	      3 -> do
		    ad <- get bh
		    return (FragmentInt ad)
	      4 -> do
		    ae <- get bh
		    return (FragmentNest ae)

instance Binary PickledPuffs where
    put bh (PickledPuffs aa) = do
	    pos <- putBits bh 0 0
	    put bh aa
	    return pos
    get bh = do
	    h <- getBits bh 0
	    case h of
	      0 -> do
		    aa <- get bh
		    return (PickledPuffs aa)



{-


showList []       = showString "[]"
showList (x:xs)   = showChar '[' . shows x . showl xs
		    where showl []     = showChar ']'
			  showl (x:xs) = showChar ',' . shows x .
					 showl xs

readList         = readParen False (\r -> [pr | ("[",s)  <- lex r,
						pr       <- readl s])
		   where readl  s = [([],t)   | ("]",t)  <- lex s] ++
				    [(x:xs,u) | (x,t)    <- reads s,
						(xs,u)   <- readl' t]
			 readl' s = [([],t)   | ("]",t)  <- lex s] ++
				    [(x:xs,v) | (",",t)  <- lex s,
						(x,u)    <- reads t,
						(xs,v)   <- readl' u]



readsPuff d input =
	      readParen (d > 9)
	      (\ inp ->
	       [((Puff aa ab) , rest) | ("Puff" , inp) <- lex inp ,
		(aa , inp) <- readsPrec 0 inp , (ab , rest) <- readsPrec 0 inp])
	      input
	      ++
	      readParen (d > 9)
	      (\ inp ->
	       [((Message aa) , rest) | ("Message" , inp) <- lex inp ,
		(aa , rest) <- readsPrec 0 inp])
	      input

showsPuff d (Puff aa ab) = showParen (d >= 10)
	      (showString "Puff" . showChar ' ' . showsPrec 0 aa . showChar ' ' .
	       showsPrec 0 ab)
showsPuff d (Message aa) = showParen (d >= 10)
	      (showString "Message" . showChar ' ' . showsPrec 0 aa)

readsFragment d input =
	      readParen (d > 9)
	      (\ inp ->
	       [((FragmentText aa) , rest) | ("FragmentText" , inp) <- lex inp ,
		(aa , rest) <- readsPrec 0 inp])
	      input
	      ++
	      readParen (d > 9)
	      (\ inp ->
	       [((FragmentData aa) , rest) | ("FragmentData" , inp) <- lex inp ,
		(aa , rest) <- readsPrec 0 inp])
	      input
	      ++
	      readParen (d > 9)
	      (\ inp ->
	       [((FragmentTime aa) , rest) | ("FragmentTime" , inp) <- lex inp ,
		(aa , rest) <- readsPrec 0 inp])
	      input
	      ++
	      readParen (d > 9)
	      (\ inp ->
	       [((FragmentInt aa) , rest) | ("FragmentInt" , inp) <- lex inp ,
		(aa , rest) <- readsPrec 0 inp])
	      input
	      ++
	      readParen (d > 9)
	      (\ inp ->
	       [((FragmentNest aa) , rest) | ("FragmentNest" , inp) <- lex inp ,
		(aa , rest) <- readsPrec 0 inp])
	      input

showsFragment d (FragmentText aa) = showParen (d >= 10)
	  (showString "FragmentText" . showChar ' ' . showsPrec 0 aa)
showsFragment d (FragmentData aa) = showParen (d >= 10)
	  (showString "FragmentData" . showChar ' ' . showsPrec 0 aa)
showsFragment d (FragmentTime aa) = showParen (d >= 10)
	  (showString "FragmentTime" . showChar ' ' . showsPrec 0 aa)
showsFragment d (FragmentInt aa) = showParen (d >= 10)
	  (showString "FragmentInt" . showChar ' ' . showsPrec 0 aa)
showsFragment d (FragmentNest aa) = showParen (d >= 10)
	  (showString "FragmentNest" . showChar ' ' . showsPrec 0 aa)

-}
